#!/usr/bin/env python3
""" oab - Outlook Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Timothy Holmes
"""

import struct

class LZXDDecompressor:
    """ Compressed version 4 full details file decompressor """
    def __init__(self, window_size):
        self.window_size = window_size
        self.window = bytearray(self.window_size)
        self.window_pos = 0
        self.r = [1, 1, 1]  # Recent offsets initialized to 1 as per LZXD
        self.first_chunk_read = False

    def _rotate_left(self, value, count, bits=16):
        """ Private method to rotate bits to the left """
        return ((value << count) | (value >> (bits - count))) & ((1 << bits) - 1)

    def _read_bit(self, bitstream):
        """ Private method to read a bit """
        if bitstream["remaining"] == 0:
            bitstream["remaining"] = 16
            if len(bitstream["buffer"]) < 2:
                raise ValueError("Unexpected EOF")
            bitstream["n"] = struct.unpack('<H', bitstream["buffer"][:2])[0]
            bitstream["buffer"] = bitstream["buffer"][2:]

        bitstream["remaining"] -= 1
        bitstream["n"] = self._rotate_left(bitstream["n"], 1)
        return bitstream["n"] & 1

    def _read_bits(self, bitstream, bits):
        """ Private method to read bits """
        result = 0
        while bits > 0:
            if bitstream["remaining"] == 0:
                bitstream["remaining"] = 16
                if len(bitstream["buffer"]) < 2:
                    raise ValueError("Unexpected EOF")
                bitstream["n"] = struct.unpack('<H', bitstream["buffer"][:2])[0]
                bitstream["buffer"] = bitstream["buffer"][2:]

            take_bits = min(bits, bitstream["remaining"])
            bitstream["remaining"] -= take_bits
            result = (result << take_bits) | (bitstream["n"] >> (16 - take_bits))
            bitstream["n"] <<= take_bits
            bits -= take_bits

        return result

    def decompress(self, compressed_data, expected_size):
        """ Public method to decompress a bitstream """
        # Initialize bitstream from the compressed data
        bitstream = {"buffer": compressed_data, "n": 0, "remaining": 0}

        # The final decompressed data
        decompressed_data = bytearray()

        while len(decompressed_data) < expected_size:
            if not self.first_chunk_read:
                # Skip the first chunk header
                _ = self._read_bit(bitstream) # e8_translation_enabled
                self.first_chunk_read = True

            # Decompress a block of data, handling matches, literals, and copy operations
            while len(decompressed_data) < expected_size:
                # Read next bit or bits from bitstream and process as per LZXD spec
                byte = self._read_bits(bitstream, 8)  # Read one byte for this example
                decompressed_data.append(byte)

                # Add to the sliding window for further matches
                self.window[self.window_pos] = byte
                self.window_pos = (self.window_pos + 1) % self.window_size

        return decompressed_data[:expected_size]



#!/usr/bin/env python3
""" oab - Outlook Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import os

import prettytable

####################################################################################################
def oab2table(data, width=None):
    """ Print our dictionary results as a PrettyTable """
    if width is None:
        if "COLUMNS" in os.environ:
            width = int(os.environ["COLUMNS"])
        else:
            width = 80

    users = prettytable.PrettyTable()
    users.field_names = ["Name", "Account", "Email", "Mobile", "Department", "Country"]
    # Distribute terminal width minus the 7 vertical borders and the 2 spaces per column
    users._max_width = {
        "Name": int(0.20 * (width - 7)) - 2,
        "Account": int(0.08 * (width - 7)) - 2,
        "Email": int(0.28 * (width - 7)) - 2,
        "Mobile": int(0.16 * (width - 7)) - 2,
        "Department": int(0.20 * (width - 7)) - 2,
        "Country": int(0.08 * (width - 7)) - 2,
    }
    users.align = "l"
    users.set_style(prettytable.SINGLE_BORDER)
    users.sortby = "Name"

    lists = prettytable.PrettyTable()
    lists.field_names = ["Name", "Account", "Email", "Members (int+ext)"]
    # Distribute terminal width minus the 5 vertical borders and the 2 spaces per column
    lists._max_width = {
        "Name": int(0.24 * (width - 5)) - 2,
        "Account": int(0.24 * (width - 5)) - 2,
        "Email": int(0.4 * (width - 5)) - 2,
        "Members (int+ext)": int(0.12 * (width - 5)) - 2,
    }
    lists.align = "l"
    lists.set_style(prettytable.SINGLE_BORDER)
    lists.sortby = "Name"

    count_users = 0
    count_lists = 0
    count_lists_without_members = 0
    count_unknown = 0

    for entry in data["address_book"]:
        record = entry["record"]

        match record["PidTagObjectType"]:
            case 6: # User
                count_users += 1
                display_name = ""
                if "PidTagDisplayName" in record:
                    display_name = record["PidTagDisplayName"]
                account = ""
                if "PidTagAddressBookDisplayNamePrintable" in record:
                    account = record["PidTagAddressBookDisplayNamePrintable"]
                elif "PidTagAccount" in record:
                    account = record["PidTagAccount"]
                email = ""
                if "PidTagSmtpAddress" in record:
                    email = record["PidTagSmtpAddress"]
                mobile = ""
                if "PidTagMobileTelephoneNumber" in record:
                    mobile = record["PidTagMobileTelephoneNumber"]
                department_name = ""
                if "PidTagDepartmentName" in record:
                    department_name = record["PidTagDepartmentName"]
                country = ""
                if "PidTagCountry" in record:
                    country = record["PidTagCountry"]

                users.add_row([display_name, account, email, mobile, department_name, country])

            case 8: # List
                count_lists += 1
                display_name = ""
                if "PidTagDisplayName" in record:
                    display_name = record["PidTagDisplayName"]
                account = ""
                if "PidTagAddressBookDisplayNamePrintable" in record:
                    account = record["PidTagAddressBookDisplayNamePrintable"]
                elif "PidTagAccount" in record:
                    account = record["PidTagAccount"]
                email = ""
                if "PidTagSmtpAddress" in record:
                    email = record["PidTagSmtpAddress"]
                members = ""
                count = 0
                if "PidTagAddressBookDistributionListMemberCount" in record:
                    count += record["PidTagAddressBookDistributionListMemberCount"]
                    members = str(record["PidTagAddressBookDistributionListMemberCount"])
                else:
                    members = "0"
                if "PidTagAddressBookDistributionListExternalMemberCount" in record:
                    count += record["PidTagAddressBookDistributionListExternalMemberCount"]
                    members += " + " + \
                        str(record["PidTagAddressBookDistributionListExternalMemberCount"])
                else:
                    members = " + 0"
                if count:
                    members = f"{str(count)} ({members})"
                else:
                    members = "0"
                    count_lists_without_members += 1

                lists.add_row([display_name, account, email, members])

            case _: # Unknown
                count_unknown += 1

    logging.info("Users count = %d", count_users)
    logging.info("Lists count = %d", count_lists)
    logging.info("Empty lists count = %d", count_lists_without_members)
    logging.info("Unknown count = %d", count_unknown)

    return users, lists



#!/usr/bin/env python3
""" oab - Outlook Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

property_tags = {
    # mandatory rgHdrAtts:
    0x6800001F: "PidTagOfflineAddressBookName",
    0x6804001E: "PidTagOfflineAddressBookDistinguishedName",
    0x68010003: "PidTagOfflineAddressBookSequence",
    0x6802001E: "PidTagOfflineAddressBookContainerGuid",
    # optional rgHdrAtts:
    0x8C98001E: "PidTagAddressBookHierarchicalRootDepartment",
    # mandatory rgOabAtts:
    0x3003001E: "PidTagEmailAddress",
    0x39fe001F: "PidTagSmtpAddress",
    # optional rgOabAtts:
    0x3001001F: "PidTagDisplayName",
    0x8C92001F: "PidTagAddressBookPhoneticDisplayName",
    0x3A00001F: "PidTagAccount",
    0x3A11001F: "PidTagSurname",
    0x8C8F001F: "PidTagAddressBookPhoneticSurname",
    0x3A06001F: "PidTagGivenName",
    0x8C8E001F: "PidTagAddressBookPhoneticGivenName",
    0x800f101F: "PidTagAddressBookProxyAddresses",
    0x3A19001F: "PidTagOfficeLocation",
    0x39000003: "PidTagDisplayType",
    0x0FFE0003: "PidTagObjectType",
    0x3A40000B: "PidTagSendRichInfo",
    0x3A08001F: "PidTagBusinessTelephoneNumber",
    0x3A0A001F: "PidTagInitials",
    0x3A29001F: "PidTagStreetAddress",
    0x3A27001F: "PidTagLocality",
    0x3A28001F: "PidTagStateOrProvince",
    0x3A2A001F: "PidTagPostalCode",
    0x3A26001F: "PidTagCountry",
    0x3A17001F: "PidTagTitle",
    0x3A16001F: "PidTagCompanyName",
    0x8C91001F: "PidTagAddressBookPhoneticCompanyName",
    0x3A30001F: "PidTagAssistant",
    0x3A18001F: "PidTagDepartmentName",
    0x8C90001F: "PidTagAddressBookPhoneticDepartmentName",
    0x8011001F: "PidTagAddressBookTargetAddress",
    0x3A09001F: "PidTagHomeTelephoneNumber",
    0x3A1B101F: "PidTagBusiness2TelephoneNumbers",
    0x3A2F101F: "PidTagHome2TelephoneNumbers",
    0x3A23001F: "PidTagPrimaryFaxNumber",
    0x3A1C001F: "PidTagMobileTelephoneNumber",
    0x3A2E001F: "PidTagAssistantTelephoneNumber",
    0x3A21001F: "PidTagPagerTelephoneNumber",
    0x3004001F: "PidTagComment",
    0x3A220102: "PidTagUserCertificate",
    0x3A701102: "PidTagUserX509Certificate",
    0x8C6A1102: "PidTagAddressBookX509Certificate",
    0x8006001E: "PidTagAddressBookHomeMessageDatabase",
    0x39FF001E: "PidTagAddressBookDisplayNamePrintable",
    0x39050003: "PidTagDisplayTypeEx",
    0x8CA00003: "PidTagAddressBookSeniorityIndex",
    0x8CDD000B: "PidTagAddressBookHierarchicalIsHierarchicalGroup",
    0x8C6D0102: "PidTagAddressBookObjectGuid",
    0x8CAC101F: "PidTagAddressBookSenderHintTranslations",
    0x806A0003: "PidTagAddressBookDeliveryContentLength",
    0x8CB5000B: "PidTagAddressBookModerationEnabled",
    0x8CE20003: "PidTagAddressBookDistributionListMemberCount",
    0x8CE30003: "PidTagAddressBookDistributionListExternalMemberCount",
    0x8009101E: "PidTagAddressBookMember",
    0x8008101E: "PidTagAddressBookIsMemberOfDistributionList",
    0x68051003: "PidTagOfflineAddressBookTruncatedProperties",
    # truncated rgOabAtts
    0x8C9E0102: "PidTagThumbnailPhoto",
    0x8CC20102: "PidTagSpokenName",
    0x8CD8000D: "PidTagAddressBookAuthorizedSenders",
    0x8CD9000D: "PidTagAddressBookUnauthorizedSenders",
    0x8073000D: "PidTagAddressBookDistributionListMemberSubmitAccepted",
    0x8CDA000D: "PidTagAddressBookDistributionListMemberSubmitRejected",
    0x8CDB000D: "PidTagAddressBookDistributionListRejectMessagesFromDLMembers",
}



#!/usr/bin/env python3
""" oab - Outlook Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import openpyxl

####################################################################################################
def oab2excel(data, filename):
    """ Output our dictionary results to a file in Excel XLSX format """
    users = []
    lists = []
    column_width = {}
    for i in range(20):
        column_width[f"users-{chr(ord('A')+i)}"] = 0
    for i in range(7):
        column_width[f"lists-{chr(ord('A')+i)}"] = 0

    for entry in data["address_book"]:
        record = entry["record"]

        match record["PidTagObjectType"]:
            case 6: # User
                display_name = ""
                if "PidTagDisplayName" in record:
                    display_name = record["PidTagDisplayName"]
                    if len(display_name) > column_width["users-A"]:
                        column_width["users-A"] = len(display_name)
                title = ""
                if "PidTagTitle" in record:
                    title = record["PidTagTitle"]
                    if len(title) > column_width["users-B"]:
                        column_width["users-B"] = len(title)
                surname = ""
                if "PidTagSurname" in record:
                    surname = record["PidTagSurname"]
                    if len(surname) > column_width["users-C"]:
                        column_width["users-C"] = len(surname)
                given_name = ""
                if "PidTagGivenName" in record:
                    given_name = record["PidTagGivenName"]
                    if len(given_name) > column_width["users-D"]:
                        column_width["users-D"] = len(given_name)
                initials = ""
                if "PidTagInitials" in record:
                    initials = record["PidTagInitials"]
                    if len(initials) > column_width["users-E"]:
                        column_width["users-E"] = len(initials)
                company_name = ""
                if "PidTagCompanyName" in record:
                    company_name = record["PidTagCompanyName"]
                    if len(company_name) > column_width["users-F"]:
                        column_width["users-F"] = len(company_name)

                account = ""
                if "PidTagAddressBookDisplayNamePrintable" in record:
                    account = record["PidTagAddressBookDisplayNamePrintable"]
                    if len(account) > column_width["users-G"]:
                        column_width["users-G"] = len(account)
                elif "PidTagAccount" in record:
                    account = record["PidTagAccount"]
                    if len(account) > column_width["users-G"]:
                        column_width["users-G"] = len(account)
                address = ""
                if "PidTagSmtpAddress" in record:
                    address = record["PidTagSmtpAddress"]
                    if len(address) > column_width["users-H"]:
                        column_width["users-H"] = len(address)
                mobile = ""
                if "PidTagMobileTelephoneNumber" in record:
                    mobile = record["PidTagMobileTelephoneNumber"]
                    if len(mobile) > column_width["users-I"]:
                        column_width["users-I"] = len(mobile)
                business_phone = ""
                if "PidTagBusinessTelephoneNumber" in record:
                    business_phone = record["PidTagBusinessTelephoneNumber"]
                if "PidTagBusiness2TelephoneNumbers" in record:
                    for number in record["PidTagBusiness2TelephoneNumbers"]:
                        business_phone += ',' + number.replace('"', "'")
                if len(business_phone) > column_width["users-J"]:
                    column_width["users-J"] = len(business_phone)
                pager = ""
                if "PidTagPagerTelephoneNumber" in record:
                    pager = record["PidTagPagerTelephoneNumber"]
                    if len(pager) > column_width["users-K"]:
                        column_width["users-K"] = len(pager)
                fax = ""
                if "PidTagPrimaryFaxNumber" in record:
                    fax = record["PidTagPrimaryFaxNumber"]
                    if len(fax) > column_width["users-L"]:
                        column_width["users-L"] = len(fax)
                home_phone = ""
                if "PidTagHomeTelephoneNumber" in record:
                    home_phone = record["PidTagHomeTelephoneNumber"]
                if "PidTagHome2TelephoneNumbers" in record:
                    for number in record["PidTagHome2TelephoneNumbers"]:
                        home_phone += ',' + number.replace('"', "'")
                if len(home_phone) > column_width["users-M"]:
                    column_width["users-M"] = len(home_phone)

                department_name = ""
                if "PidTagDepartmentName" in record:
                    department_name = record["PidTagDepartmentName"]
                    if len(department_name) > column_width["users-N"]:
                        column_width["users-N"] = len(department_name)
                office_location = ""
                if "PidTagOfficeLocation" in record:
                    office_location = record["PidTagOfficeLocation"]
                    if len(office_location) > column_width["users-O"]:
                        column_width["users-O"] = len(office_location)
                street_address = ""
                if "PidTagStreetAddress" in record:
                    street_address = record["PidTagStreetAddress"]
                    if len(street_address) > column_width["users-P"]:
                        column_width["users-P"] = len(street_address)
                postal_code = ""
                if "PidTagPostalCode" in record:
                    postal_code = record["PidTagPostalCode"]
                    if len(postal_code) > column_width["users-Q"]:
                        column_width["users-Q"] = len(postal_code)
                locality = ""
                if "PidTagLocality" in record:
                    locality = record["PidTagLocality"]
                    if len(locality) > column_width["users-R"]:
                        column_width["users-R"] = len(locality)
                state = ""
                if "PidTagStateOrProvince" in record:
                    state = record["PidTagStateOrProvince"]
                    if len(state) > column_width["users-S"]:
                        column_width["users-S"] = len(state)
                country = ""
                if "PidTagCountry" in record:
                    country = record["PidTagCountry"]
                    if len(country) > column_width["users-T"]:
                        column_width["users-T"] = len(country)

                users.append(
                    {
                        "display_name": display_name,
                        "title": title,
                        "surname": surname,
                        "given_name": given_name,
                        "initials": initials,
                        "company_name": company_name,
                        "account": account,
                        "address": address,
                        "mobile": mobile,
                        "business_phone": business_phone,
                        "pager": pager,
                        "fax": fax,
                        "home_phone": home_phone,
                        "department_name": department_name,
                        "office_location": office_location,
                        "street_address": street_address,
                        "postal_code": postal_code,
                        "locality": locality,
                        "state": state,
                        "country": country,
                    }
                )

            case 8: # List
                display_name = ""
                if "PidTagDisplayName" in record:
                    display_name = record["PidTagDisplayName"]
                    if len(display_name) > column_width["lists-A"]:
                        column_width["lists-A"] = len(display_name)
                account = ""
                if "PidTagAddressBookDisplayNamePrintable" in record:
                    account = record["PidTagAddressBookDisplayNamePrintable"]
                    if len(account) > column_width["lists-B"]:
                        column_width["lists-B"] = len(account)
                elif "PidTagAccount" in record:
                    account = record["PidTagAccount"]
                    if len(account) > column_width["lists-B"]:
                        column_width["lists-B"] = len(account)
                address = ""
                if "PidTagSmtpAddress" in record:
                    address = record["PidTagSmtpAddress"]
                    if len(address) > column_width["lists-C"]:
                        column_width["lists-C"] = len(address)
                members = 0
                internal_members = 0
                external_members = 0
                if "PidTagAddressBookDistributionListMemberCount" in record:
                    internal_members = record["PidTagAddressBookDistributionListMemberCount"]
                    if len(str(internal_members)) > column_width["lists-E"]:
                        column_width["lists-E"] = len(str(internal_members))
                if "PidTagAddressBookDistributionListExternalMemberCount" in record:
                    external_members = \
                        record["PidTagAddressBookDistributionListExternalMemberCount"]
                    if len(str(external_members)) > column_width["lists-F"]:
                        column_width["lists-F"] = len(str(external_members))
                members = internal_members + external_members
                if len(str(members)) > column_width["lists-D"]:
                    column_width["lists-D"] = len(str(members))
                comment = ""
                if "PidTagComment" in record:
                    comment = record["PidTagComment"]
                    if len(comment) > column_width["lists-G"]:
                        column_width["lists-G"] = len(comment)

                lists.append(
                    {
                        "display_name": display_name,
                        "account": account,
                        "address": address,
                        "members": members,
                        "internal_members": internal_members,
                        "external_members": external_members,
                        "comment": comment,
                    }
                )

            case _: # Unknown
                pass


    workbook = openpyxl.Workbook()

    # First tab for users
    sheet = workbook.active
    sheet.title = "Users"
    sheet.page_setup.orientation = sheet.ORIENTATION_LANDSCAPE
    sheet.page_setup.fitToWidth = 1
    sheet["A1"] = "Name"
    sheet["B1"] = "Title"
    sheet["C1"] = "Surname"
    sheet["D1"] = "Given name"
    sheet["E1"] = "Initials"
    sheet["F1"] = "Company"
    sheet["G1"] = "Account"
    sheet["H1"] = "Email"
    sheet["I1"] = "Mobile"
    sheet["J1"] = "Business phone"
    sheet["K1"] = "Pager"
    sheet["L1"] = "Fax"
    sheet["M1"] = "Home phone"
    sheet["N1"] = "Department"
    sheet["O1"] = "Office location"
    sheet["P1"] = "Street address"
    sheet["Q1"] = "Postal code"
    sheet["R1"] = "Locality"
    sheet["S1"] = "State or province"
    sheet["T1"] = "Country"
    for i in range(20):
        sheet.column_dimensions[f"{chr(ord('A')+i)}"].width = \
            column_width[f"users-{chr(ord('A')+i)}"]
        sheet[f"{chr(ord('A')+i)}1"].font = openpyxl.styles.Font(bold=True)
    sheet.freeze_panes = "A2"

    for i, value in enumerate(users):
        sheet[f"A{i+2}"] = value["display_name"]
        sheet[f"B{i+2}"] = value["title"]
        sheet[f"C{i+2}"] = value["surname"]
        sheet[f"D{i+2}"] = value["given_name"]
        sheet[f"E{i+2}"] = value["initials"]
        sheet[f"F{i+2}"] = value["company_name"]
        sheet[f"G{i+2}"] = value["account"]
        sheet[f"H{i+2}"] = value["address"]
        sheet[f"I{i+2}"] = value["mobile"]
        sheet[f"J{i+2}"] = value["business_phone"]
        sheet[f"K{i+2}"] = value["pager"]
        sheet[f"L{i+2}"] = value["fax"]
        sheet[f"M{i+2}"] = value["home_phone"]
        sheet[f"N{i+2}"] = value["department_name"]
        sheet[f"O{i+2}"] = value["office_location"]
        sheet[f"P{i+2}"] = value["street_address"]
        sheet[f"Q{i+2}"] = value["postal_code"]
        sheet[f"R{i+2}"] = value["locality"]
        sheet[f"S{i+2}"] = value["state"]
        sheet[f"T{i+2}"] = value["country"]
        for j in range(20):
            sheet[f"{chr(ord('A')+j)}{i+2}"].alignment = \
                openpyxl.styles.Alignment(vertical="top", wrapText=True)

    # Second tab for lists
    workbook.create_sheet(index=1, title="Lists")
    sheet = workbook.get_sheet_by_name("Lists")
    sheet.page_setup.orientation = sheet.ORIENTATION_LANDSCAPE
    sheet.page_setup.fitToWidth = 1
    sheet["A1"] = "Name"
    sheet["B1"] = "Account"
    sheet["C1"] = "Email"
    sheet["D1"] = "Members"
    sheet["E1"] = "Internal members"
    sheet["F1"] = "External members"
    sheet["G1"] = "Comment"
    for i in range(7):
        sheet.column_dimensions[f"{chr(ord('A')+i)}"].width = \
            column_width[f"lists-{chr(ord('A')+i)}"]
        sheet[f"{chr(ord('A')+i)}1"].font = openpyxl.styles.Font(bold=True)
    sheet.freeze_panes = "A2"

    for i, value in enumerate(lists):
        sheet[f"A{i+2}"] = value["display_name"]
        sheet[f"B{i+2}"] = value["account"]
        sheet[f"C{i+2}"] = value["address"]
        sheet[f"D{i+2}"] = value["members"]
        sheet[f"E{i+2}"] = value["internal_members"]
        sheet[f"F{i+2}"] = value["external_members"]
        sheet[f"G{i+2}"] = value["comment"]
        for j in range(7):
            sheet[f"{chr(ord('A')+j)}{i+2}"].alignment = \
                openpyxl.styles.Alignment(vertical="top", wrapText=True)

    workbook.save(filename)
